/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.Image;
import java.awt.Insets;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.util.Collection;
import java.util.Iterator;

import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JTabbedPane;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import net.sf.dc.console.components.ComponentFactory;
import net.sf.dc.console.components.DCFrame;
import net.sf.dc.console.components.Layout;
import net.sf.dc.console.dialogs.AboutDialog;
import net.sf.dc.console.note.NotePanel;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DCModule;
import net.sf.dc.core.modules.DCModules;
import net.sf.dc.core.objects.DCFieldDefinition;
import net.sf.dc.db.Database;
import net.sf.dc.helpsystem.JavaHelp;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DCLabels;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.settings.ReportSettings;
import net.sf.dc.wf.requests.IRequest;
import net.sf.dc.wf.requests.RequestCollection;
import net.sf.ext.jflexsett.settings.ui.SettingsView;
import net.sf.ext.util.Resources;
import net.sf.ext.util.Utilities;

/**
 * Holder of all the UI panels. The panels are shown as tabs.
 * By using this frame background colors can be set and requests can be past
 * to the panels by using the @see RequestorCollection class.
 * At the top of the frame a menubar and a toolbar are shown. 
 * From these actions can be started
 * 
 * NOTE: I have specifically chosen to load all the panels at once. 
 *       Of course one can say this is heavy on the resources, but.. 
 *       There are just a couple of panels. Switching between modules
 *       should be smooth. Constantly needing to rebuild panels and 
 *       other UI stuff would make the application slow
 *
 * @author Robert Jan van der Waals
 * @since 1.3
 * @version 0.50
 */
public class MainFrame extends DCFrame {
    
	private static final long serialVersionUID = 9L;
	public static final int _SEARCHTAB = 0;
    public static final int _INSERTTAB = 1;
    
    private final JMenuBar menuBar = new JMenuBar();
    private final DCToolBar toolBar = new DCToolBar();
    
    private final ModuleListPanel moduleListPanel = new ModuleListPanel();
    private final JTabbedPane tabbedPane = ComponentFactory.getTabbedPane();
    private final NotePanel notePanel = new NotePanel();

    private SettingsView settingsView;
    private DCModule activeModule = null;

    /**
     * Initializes all the panels and puts them in a tabbed pane. The menubar
     * is build and its actions specified. The background color for all UI
     * components is set
     */
    public MainFrame() {
        super(null);
        
        setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE); 
        
        this.getContentPane().setLayout(Layout.getGBL());
        
        // Initializes the settings editor
        createSettingsView();

        toggleModuleList(DcSettings.getValueAsBoolean(Repository.settings.stShowModuleList));
        
        // listen for tab selections
        tabbedPane.getModel().addChangeListener(new TabIndexListener());
        
        try {
            setTitle(Resources.getText(DCLabels.lblMainTitle));
            setResizable(false);
            addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    finish(true);
                }
            });

            Image image = IconLibrary._icoMain.getImage();
            setIconImage(image);
            image.flush();

            int module = DcSettings.getValueAsInt(Repository.settings.stModule);
            changeModule(module);

            getContentPane().add(toolBar, Layout.getGBC( 0, 0, 2, 1, 1.0, 1.0
                    ,GridBagConstraints.NORTHWEST, GridBagConstraints.VERTICAL,
                     new Insets(0, 0, 0, 0), 0, 0));

            getContentPane().add(moduleListPanel, Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
                                ,GridBagConstraints.NORTHWEST, GridBagConstraints.VERTICAL,
                                 new Insets(0, 0, 0, 0), 0, 0));
            getContentPane().add(tabbedPane, Layout.getGBC( 1, 1, 1, 1, 20.0, 20.0
                                 ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                                 new Insets(0, 0, 0, 0), 0, 0));
            
            tabbedPane.setPreferredSize(DcSettings.getValueAsDimension(Repository.settings.stTabbedPaneSize));
            pack();
            tabbedPane.setSize(DcSettings.getValueAsDimension(Repository.settings.stTabbedPaneSize));
            
            setHiddenFields();
        } catch (Error err) {
            Messages.add(err);
        }
        this.setBackgroundColor(
                DcSettings.getValueAsColor(Repository.settings.stColor));
        
        this.setRequiredFields();
        this.setResizable(true);
        this.setSize(DcSettings.getValueAsDimension(Repository.settings.stMainViewSize));
        Dimension location = Utilities.getCenteredWindowLocation(getSize());
        setLocation(location.width, location.height);
    }
    
    public void setVisible(boolean b) {
        super.setVisible(b);
        
        getContentPane().setBackground(menuBar.getBackground());
    }
    
    /**
     * Set the selected tab
     * @param index tab index
     */
    public void setSelectedTab(int index) {
    	tabbedPane.setSelectedIndex(index);
    }
    
    /**
     * Returns the currently selected tab
     */
    public int getSelectedTab() {
        return tabbedPane.getSelectedIndex();
    }

    /**
     * Displays the settingsview
     */
    public void showSettingsView() {
    	settingsView.setVisible(true);
    }

    /**
     * Displays the about dialog
     */
    public void showAboutDialog() {
        new AboutDialog(this).setVisible(true);
    }    
    
    /**
     * Returns the blocknote panel
     * @return
     */
    public NotePanel getNotePanel() {
    	return notePanel;
    }
    
    private void toggleModuleList(boolean enable) {
        moduleListPanel.setVisible(enable);
    }
    
    public void toggleModuleList() {
    	moduleListPanel.setVisible(!moduleListPanel.isVisible());
        DcSettings.setValue(Repository.settings.stShowModuleList, 
                            new Boolean(moduleListPanel.isVisible()));
    }
    
    public DCToolBar getToolbar() {
    	return toolBar;
    }
    
    /**
     * Applies the required fields for all modules. A required
     * field will mostly be shown as a red header in the table.
     */
    public void setRequiredFields() {
        for (Iterator iter = DCModules.getModules(); iter.hasNext(); ) {
            DCModule module = (DCModule) iter.next();
            if (module.getSearchPanel() != null) {
                module.getSearchPanel().setRequiredFields();
            }
            
            if (module.getInsertPanel() != null) {
            	module.getInsertPanel().setRequiredFields();
            }
        }
        this.repaint();
    }
    
    /**
     * Removes all the fields from the table which are not selected as 
     * visible fields. These fields are solely removed from the table,
     * not from the tablemodel. They will be updated and used in validations.
     */
    public static void setHiddenFields() {
        for (Iterator iter = DCModules.getModules(); iter.hasNext(); ) {
            DCModule module = (DCModule) iter.next();
            int[] hiddenFields = DcSettings.getValueAsIntegerArray(
                                    module.getHiddenFieldsSettingsKey());
            
            DCFieldDefinition[] fields = module.getFieldDefinitions();
            for (int j = 0; j < fields.length; j++) {
                fields[j].setHidden(false);
            }
            
            for (int i = 0; i < hiddenFields.length; i++) {
                fields[hiddenFields[i]].setHidden(true); 
            }
            
            if (module.getSearchPanel() != null) {
                module.getSearchPanel().removeHiddenFields(hiddenFields);
            }
            
            if (module.getInsertPanel() != null) {
                module.getInsertPanel().removeHiddenFields(hiddenFields);
            }
        }
    }
    
    /**
     * Applies the background color to all panels and their components
     * @color   background color
     */
    public void setBackgroundColor(Color color) {
        for (Iterator iter = DCModules.getModules(); iter.hasNext(); ) {
            DCModule module = (DCModule) iter.next();
            
            if (module.getSearchPanel() != null) {
                module.getSearchPanel().setBackground(color);
            }   
            
            if (module.getInsertPanel() != null) {
            	module.getInsertPanel().setBackground(color);
            }
        }            
        getContentPane().setBackground(menuBar.getBackground());
        tabbedPane.setBackground(color);
        settingsView.setBackgroundColor(color);            
        notePanel.setBackground(color);
    }

    /**
     * Sets the module. Only the tabs for this module are shown. The module
     * is selected from the module combobox from the menubar
     */
    public void changeModule(int index) {
        activeModule = DCModules.getModule(index);
        
        DcSettings.setValue(Repository.settings.stModule, new Integer(index));
        
        moduleListPanel.setSelectedIndex(index);
        setJMenuBar(null);
        setJMenuBar(activeModule.getMenuBar());
        
        Dimension size = tabbedPane.getSize();
        
        tabbedPane.setPreferredSize(size);
        tabbedPane.removeAll();
        
        DCPanel[] panels = activeModule.getPanels();
        for (int i = 0; i < panels.length; i++) {
            tabbedPane.addTab(panels[i].getTitle(), 
                              panels[i].getIcon(), 
							  panels[i]);
        }
        
        tabbedPane.addTab(Resources.getText(DCLabels.lblTabNote), 
                          IconLibrary._icoNote, 
                          notePanel);
        
        tabbedPane.setSize(size);
        tabbedPane.setPreferredSize(size);
    }
    
    /**
     * Closes the application. If running in client <> server mode, a question 
     * will popup to close the server as well.
     */
    public void finish(boolean closeDB) {
        Messages.add(
            Resources.getText(DCMessages.msgApplicationStops),
            Messages._STANDARD);

        Database.close();

        DcSettings.setValue(Repository.settings.stTabbedPaneSize, tabbedPane.getSize());
        DcSettings.setValue(Repository.settings.stMainViewSize, getSize());
        
        DcSettings.save();
        ReportSettings.save();
        clearCache();
        
        System.exit(0);
    }
    /**
     * Passes requests to the correct module (read, the UI component of
     * the module). 
     * 
     * @param requestors a collection of requests (actions) to be passed 
     *                   to the target
     * @param data data which should be passed to the target. Can be null
     * @param qryWasSuccess Indicates if the query was executed succesfully.
     *                      Unsuccessfull queries can still result in the
     *                      actions being executed (if specified)
     */
    public void displayResults( RequestCollection requests, 
                                Collection objects, 
                                boolean qryWasSuccess) {
        for (Iterator iter = requests.iterator(); iter.hasNext(); ) {
        	IRequest request = (IRequest) iter.next();
            
            if (qryWasSuccess || (!qryWasSuccess && request.getExecuteOnFail())) {
                request.execute(objects);
            }
        }
    }
    
    /**
     * Removes the cache from the cache directory. The cache directory
     * itself will always remain after its creation.
     * Cache consists mostly out of pictures retrieved from the database
     */
    private void clearCache() {
        try {
            String sCacheDir = Utilities.getCacheDirectory();
            File file = new File(sCacheDir);
            if (file.exists() && file.isDirectory()) {
                String[] files  = file.list();
                
                if (files != null) {
                    for (int i = 0; i < files.length; i++) {
                        File cacheFl = new File (sCacheDir + files[i]);
                        cacheFl.delete();
                    }
                }
            }
        } catch (Exception exp) {
            Messages.add(Resources.getText(DCMessages.msgCouldNotClearCache),
                                Messages._ERROR);
            Messages.add(exp);
        }
    }
    
    private void createSettingsView() {
        settingsView = DcSettings.getView(this);
    }   
    
    private class TabIndexListener implements ChangeListener {
        public void stateChanged(ChangeEvent arg0) {
            int index = tabbedPane.getSelectedIndex();
            if (index > -1) {
            	DCPanel panel = (DCPanel) tabbedPane.getComponent(index);
            	JavaHelp.setHelpIndex(panel.getHelpIndex());
            }
	    }
    }
}