/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;

import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableColumn;

import net.sf.dc.console.MainFrame;
import net.sf.dc.console.components.renderers.ComboBoxTableCellRenderer;
import net.sf.dc.console.components.renderers.DCTableCellRenderer;
import net.sf.dc.console.components.renderers.DCTableHeaderRenderer;
import net.sf.dc.console.components.renderers.DCTableHeaderRendererRequired;
import net.sf.dc.console.components.renderers.DCTableRowIndexCellRenderer;
import net.sf.dc.console.components.renderers.NumberTableCellRenderer;
import net.sf.dc.console.components.renderers.PictureTableCellRenderer;
import net.sf.dc.console.components.renderers.RatingTableCellRenderer;
import net.sf.dc.console.components.renderers.TimeFieldTableCellRenderer;
import net.sf.dc.console.dialogs.FieldSelectorDialog;
import net.sf.dc.console.dialogs.MessageBox;
import net.sf.dc.console.dialogs.UpdateAllDialog;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DCModule;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.settings.DcSettings;
import net.sf.ext.util.Resources;
import net.sf.ext.util.StringUtils;

/**
 * Table implementation with caching of changed values.
 * Can either be initialized as a simple table or a table for module (with
 * all the correct components and headers)
 *
 * @author Robert Jan van der Waals
 * @since 1.4
 * @version 1.9
 */
public class DCTable extends JTable  {
    
    private final DCModule module;
    private final DCTableModel tableModel;
    private final Hashtable cache = new Hashtable();
    
    private final TableValueChangedAction tableChangeListener = 
        new TableValueChangedAction();    
    
    private final boolean caching;
    private final boolean readonly;
    private boolean ignoreEdit = false;
    
    private ArrayList alHiddenColumns = new ArrayList();
    private ArrayList alColumns = new ArrayList();
    
    private TableColumn rowIndexColumn;

    /**
     * Initializes a very simple table
     */
    public DCTable(boolean readonly, boolean caching) {
        super(new DCTableModel());
        this.readonly = readonly;
        createRowIndexColumn();
        addRowIndexColumn();
        setProperties();
        module = null;
        tableModel = (DCTableModel) getModel();
        this.caching = caching;
        readonly = false;
    }

    /**
     * Initializes a table for a specific module
     * @param module the mocule for which this table is created
     * @param readonly allow / block changes 
     */
    public DCTable(DCModule module, boolean readonly, boolean caching) {
        super(new DCTableModel(module));
        tableModel = (DCTableModel) getModel();

        this.caching = caching;
        this.module = module;
        this.readonly = readonly;
        
        buildTable();
        setListeningForChanges(true);
    }

    public void addObject(DCObject dco) {
        addObject(dco, true);
    }
    
    public void addObject(DCObject dco, boolean setSelected) {
        Object[] values = dco.getValues();
        tableModel.addRow(values);
        if (setSelected) {
        	setSelectedRow(getRowCount() - 1);
        }
    }
    
    public void addObjects(Collection objects) {
        for (Iterator iter = objects.iterator(); iter.hasNext();) {
            DCObject dco = (DCObject) iter.next();
            addObject(dco, false);
        }
        setSelectedRow(getRowCount() - 1);
    }
    
    /**
     * Updates the row with the given ID with the data (values)
     * @param ID the ID (Product Object ID)
     * @param values the new values
     */
    public void updateRow(String ID, DCObject dco) {
        cancelEdit();

        int row = getRowNumberWithID(ID);
        if (row != -1) {
            for (int j = 0; j < dco.getFields().length; j++) {
                tableModel.setValueAt(dco.getValue(j), row, j);
                addRowToCache(row, j);
            }
        }
    }     
    
    /**
     * Indicates if the table data can be edited
     */
    public boolean isReadOnly() {
    	return readonly;
    }
    
    /** 
     * Indicates if changes have been made to the data of the table
     */
    public boolean noChanges() {
        return cache.size() == 0;
    }    
    
    /**
     * Ignores the current edit action of a cell
     * @param b
     */
    public void ignoreEdit(boolean b) {
    	ignoreEdit = b;
    }
    
    /**
     * (Re-) applies all headers
     */
    public void applyHeaders() {
    	removeRowIndexColumn();
        for (Enumeration enumCols = getColumnModel().getColumns(); enumCols.hasMoreElements(); ) {
            TableColumn column = (TableColumn) enumCols.nextElement();
            column.setHeaderRenderer(getTableHeaderRenderer());
            alColumns.add(column);
        }
        addRowIndexColumn();
    }
    
    public void moveRowDown() {
        int row = getSelectedRow();
        
        if (row > -1) {
            int total = getRowCount();
            if (row < total -1) {
                int destination = row + 1;
                tableModel.moveRow(row, row, destination);
                setSelectedRow(destination);
            }
        } else {
            new MessageBox("No row selected. First select a row to move", MessageBox._WARNING);
        }
    }
    
    public void moveRowUp() {
        int row = getSelectedRow();
        
        if (row > -1) {
            if (row != 0) {
                int destination = row - 1;
                tableModel.moveRow(row, row, destination);
                setSelectedRow(destination);
            }
        } else {
            new MessageBox("No row selected. First select a row to move", MessageBox._WARNING);
        }
    }
    
    /**
     * All columns to be removed are still visible in the table. The columns 
     * are retrieved, stored in a Vector and then removed from the view
     * @param hiddenColumns The model indices of the columns
     */
    public void removeHiddenFields(int[] hiddenColumns) {
        cancelEdit();
        resetTable();
        
        alHiddenColumns.clear();
        removeRowIndexColumn();
        
        for (int i = 0; i < hiddenColumns.length; i++) {
            int tableIndex = convertColumnIndexToView(hiddenColumns[i]);
            TableColumn column = getColumnModel().getColumn(tableIndex);
            alHiddenColumns.add(column); 
            removeColumn(column);
        }
        addRowIndexColumn();
        setRequiredFields();
    }
    
    /**
     * Clears the cached objects
     */
    public void clearCache() {
    	cache.clear();
    }
    
    /**
     * Removes the object with the given ID from the cache (unmark as changed)
     * @param sID object ID
     */
    public void removeFromCache(String sID) {
    	cache.remove(sID);
    }
    
    /**
     * Returns all Data Crow Objects currently persisting in the (UI) table
     */
    public DCObject[] getDCObjects() {
        DCObject[] objects = new DCObject[tableModel.getRowCount()];
        for (int i = 0; i < objects.length; i++) {
            objects[i] = getDCObject(i);
        }
        return objects;
    }   
    
    private boolean isCached(String id) {
    	return cache.containsKey(id);
    }
    
    /**
     * Returns the object with the given ID from the cache
     * @param sID Unique object ID
     */
    public DCObject getDCObject(int row) {
        String id = getValueAt(row, 0, true).toString();
        boolean isCached = isCached(id);
        
        return isCached ? (DCObject) cache.get(id) :
                          getObjectForRow(row);
    }
    
    private DCObject getObjectForRow(int index) {
        cancelEdit();
        DCObject dco = null;
        
        if (index == -1) {
        	new MessageBox(Resources.getText(DCMessages.msgNoRowsSelected), 
                           MessageBox._WARNING);
        } else {
        	Object[] row = new Object[tableModel.getColumnCount()];
        	for (int j = 0; j < row.length; j++) {
        		Object o = getValueAt(index, j, true);
        		row[j] = (o == null ? "" : o);
        	}
            dco = module.getDCObject();
            dco.setValues(row);
        }
        return dco;
    }
    
    /**
     * Retrieve all changed objects. Use with care, this could be heavy on the
     * system when many rows have been changed. Consider using the method
     * getChangedObjectsIndices
     */
    public DCObject[] getChangedObjects() {
        DCObject[] objects = new DCObject[cache.size()];
        int counter = 0;
        for (Enumeration enumKeys = cache.keys(); enumKeys.hasMoreElements(); counter++) {   
            String sID = enumKeys.nextElement().toString();
            DCObject obj = (DCObject) cache.get(sID);
            objects[counter] = obj;
        }
        return objects;
    }    
    
    /**
     * Retrieves the rownumbers of all changed objects
     */
    public int[] getChangedObjectsIndices() {
        cancelEdit();
        int[] rows = new int[cache.size()];
        int counter = 0;
    	for (Enumeration enumKeys = cache.keys(); enumKeys.hasMoreElements(); counter++) {   
            String sID = enumKeys.nextElement().toString();
            int row = getRowNumberWithID(sID);
            rows[counter] = row;
        }
        return rows;
    }

    /**
     * Gets a value from the model (hidden or visible) 
     * @param row row index
     * @param col column index
     */
    public Object getValueAt(int row, int col, boolean hidden) {
        if (hidden) {
        	return tableModel.getValueAt(row, col);
        } else {
        	return super.getValueAt(row, col);
        }
    }
    
    /**
     * Clear the table of all values. Clear the cache as well.
     */
    public void clearTable() {
        cache.clear();
        cancelEdit();

        setListeningForChanges(false);
        tableModel.setRowCount(0);
        setListeningForChanges(true);
    }
    
    /**
     * Returns the row number of the item with the given ID
     * (product object ID; ie. the value of Software._ID)
     * @param ID
     */
    public int getRowNumberWithID(String ID) {
        cancelEdit();
        String sID = "";
        for (int i = 0; i < tableModel.getRowCount(); i++) {
            sID = tableModel.getValueAt(i, 0).toString();
            if (sID.equals(ID)) {
                return i;
            }
        }
        return -1;
    }  
    
    /**
     * Sets the amount of columns to be displayed
     * @param count columns amount
     */
    public void setColumnCount(int count) {
    	tableModel.setColumnCount(count);
    }
    
    /**
     * Remove the rows from the table with the given indices
     */
    public void removeRows(int[] rows) {
        cancelEdit();
        for (int i = rows.length -1; i > -1; i--) {
        	if (caching) {
    			removeFromCache(getValueAt(rows[i], 0, true).toString());
        	}
            tableModel.removeRow(rows[i]);
        }
    }   
    
    /**
     * Removes a row from the UI table
     * @param row row index
     */
    public void removeRow(int row) {
        tableModel.removeRow(row);
    }    
    
    /**
     * Sets the amount of rows to be displayed
     * @param count rows amount
     */
    public void setRowCount(int count) {
        tableModel.setRowCount(count);
    }    

    /**
     * Tries to programmatically set the selected row in the table.
     * @param row index
     */
    public void setSelectedRow(int row) {
        try {
            if (getSelectedRow() > -1) {
                removeColumnSelectionInterval(0, getColumnCount() -1);
                removeRowSelectionInterval(getSelectedRow(), getSelectedRow());
            }
                
            addRowSelectionInterval(row, row);
            addColumnSelectionInterval(0, getColumnCount() -1);
        } catch (Exception ignore) {}
    }
    
    /**
     * Updates a row at the given rownumber. If indicated this will
     * update existing data.
     * @param rowNumber UI row to be updated
     * @param rowData the data to be written to the row
     * @param overWrite indicates of existing data should be overwritten
     */
    public void updateRowAt(int row,
                            DCObject dco,
                            boolean overwrite,
                            boolean allowDeletes) {
        cancelEdit();
        Object[] values = dco.getValues();
        
        for (int i = 1; i < values.length; i++) {
            Object oNew = values[i];
            Object oExist = tableModel.getValueAt(row, i);
            
            if (overwrite && ((oNew != null && !oNew.equals("")) || allowDeletes)) {
                tableModel.setValueAt(oNew, row, i);
            } else if (oExist == null || oExist.equals("")) {
                tableModel.setValueAt(oNew, row, i);
            }
        }
    } 
    
    public void addRow(Object[] row) {
    	tableModel.addRow(row);
    }
    
    /**
     * Add a changed row to the cache. Can also be used to force save items
     * (add a unchanged row to the cache) 
     * @param row changed row number
     * @param column changed column number
     */
    public void addRowToCache(int row, int column) {
        setListeningForChanges(false);
        
        if (row != -1 && column != -1) {
            String sID = tableModel.getValueAt(row, 0).toString();
            
            if (sID != null && !sID.equals("")) {
                DCObject dco;
                if (!cache.containsKey(sID)) {
                    dco = getObjectForRow(row);
                    dco.undoChanges();
                } else {
                    dco = (DCObject) cache.get(sID);
                }
                dco.setValue(column, tableModel.getValueAt(row, column));
                cache.put(sID, dco);
            }
        }        
        setListeningForChanges(true);
    }      
    
    /**
     * Fills the table with the specified values
     * @param data list containing the rows of data to be displayed
     */
    public void fillTable(Collection objects) {
        clearTable();

        setListeningForChanges(false);
        
        if (objects != null) {
            for (Iterator iter = objects.iterator(); iter.hasNext(); ) {
                DCObject dco = (DCObject) iter.next(); 
            	addObject(dco, false);
            }
        }
        
        setSelectedRow(getRowCount() -1);
        setListeningForChanges(true);
    }    
    
    /**
     * Creates a visual representation (different header) 
     * of the required fields
     * @param o data crow object (used as a reference)
     */
    public void setRequiredFields() {
        DCObject dco = module.getDCObject();
        int counter = 0;
        removeRowIndexColumn();
        for (Iterator iter = alColumns.iterator(); iter.hasNext(); ) {
            TableColumn column = (TableColumn) iter.next();
            if (dco.isRequired(counter)) {
                column.setHeaderRenderer(getRequiredTableHeaderRenderer());
            } else {
                column.setHeaderRenderer(getTableHeaderRenderer());
            }
            counter++;
        }
        addRowIndexColumn();
    }    
    
    /**
     * Tries to apply the current editing of the table cell.
     * If the current value cannot be applied, the editing is simply
     * cancelled (changes are discarded) 
     */
    public void cancelEdit() {
        int selectedRow = getSelectedRow();
        if (selectedRow != -1) {
            for (int i = 0; i < getColumnCount(); i++) {
                boolean success = false;
                try {
                    getCellEditor(selectedRow, i).stopCellEditing();
                    success = true;
                } catch (Exception ignore) {}
                
                if (!success) {
                    getCellEditor(selectedRow, i).cancelCellEditing();
                }
            }
        }
    }   

    //*************************************************************************
    //Dialogs
    //*************************************************************************
    
    /**
     * Opens the fields selector dialog and applies the hidden fields on the
     * table. Also the hidden fields setting is set for the current module.
     */
    public void openFieldSelector() {
        ArrayList alVisibleColumns = new ArrayList();
        removeRowIndexColumn();
        for (Enumeration enumCols = getColumnModel().getColumns(); enumCols.hasMoreElements(); ) {
            alVisibleColumns.add(enumCols.nextElement());
        }
        FieldSelectorDialog dialog = new FieldSelectorDialog(alVisibleColumns, 
                                                             alHiddenColumns);
        addRowIndexColumn();
        dialog.setVisible(true);
        if (FieldSelectorDialog.isSuccessful()) {
            //alVisibleColumns = dialog.getVisibleFields(); 
            alHiddenColumns = (ArrayList) dialog.getHiddenFields();

            // get the hidden columns table model indices by comparing the
            // columns of the table model with the table              
            int[] modelIndices = new int[alHiddenColumns.size()];
            int counter = 0;
            for (Iterator iter = alHiddenColumns.iterator(); iter.hasNext(); ) {
                TableColumn column = (TableColumn) iter.next();
                modelIndices[counter++] = column.getModelIndex();
            }            
            DcSettings.setValue(module.getHiddenFieldsSettingsKey(), modelIndices);
            MainFrame.setHiddenFields();
        }
    }    
    
    /**
     * Opens the Update All dialog
     */
    public void openUpdateAllDialog() {
        UpdateAllDialog dialog = new UpdateAllDialog(
                                        module.getDCObject(), 
                                        DataCrow.mainFrame);
        // if the update was a success
        if (dialog.isSuccess()) {
            // retrieve the data which will be applied on the table
            Object[] values = dialog.getData();
            setListeningForChanges(false);
            if (dialog.isUpdateSelectedItemsOnly()) {
            	int[] rows = getSelectedRows();
                for (int i = 0; i < rows.length; i++) {
                    for (int j = 0; j < values.length; j++) {
                        // only update a field with a 'real' / not empty value
                        if (!isEmptyValue(values[j])) {
                            tableModel.setValueAt(values[j], rows[i], j);
                            addRowToCache(rows[i], j);
                        }
                    }
                }
            } else {
                int rows = tableModel.getRowCount();
                for (int i = 0; i < rows; i++) {
                    for (int j = 0; j < values.length; j++) {
                        // only update a field with a 'real' / not empty value
                        if (!isEmptyValue(values[j])) {
                            tableModel.setValueAt(values[j], i, j);
                            addRowToCache(i, j);
                        }
                    }
                }
            }
            setListeningForChanges(true);
        }
    }  
    
    private boolean isEmptyValue(Object o) {
        return (   o == null
        		|| StringUtils.isEmpty(o.toString())
				|| o.toString().equals("-1")
				|| o.toString().equals("0"));
    }

    //*************************************************************************
    //Private methods and classes
    //*************************************************************************
    private void buildTable() {
        DCObject dco = module.getDCObject();
        
        String[] headers = dco.getLabels();
        int[] maxLength = dco.getMaxFieldLengths();
        int[] fieldTypes = dco.getFieldTypes();

        tableModel.setColumnCount(fieldTypes.length);

        JTextField text = ComponentFactory.getTextField();
        for (int i = 0; i < fieldTypes.length; i++) {
            TableColumn columnNew = getColumnModel().getColumn(i);
            columnNew.setHeaderValue(headers[i]);

            JComponent comp = ComponentFactory.getComponent(fieldTypes[i],
                                           headers[i],
                                           maxLength[i]);
            comp.setAutoscrolls(false);
            comp.setBorder(null);
            comp.setDoubleBuffered(true);
            comp.setIgnoreRepaint(true);
            comp.setVerifyInputWhenFocusTarget(false);
            
            if (fieldTypes[i] == ComponentFactory._LONGTEXTFIELD) {
                comp = ComponentFactory.getComponent(
                                ComponentFactory._SHORTTEXTFIELD,
                                headers[i],
                                maxLength[i]);
            }
            
            if (i == 0) {
                text.setEditable(false);
                text.setEnabled(false);
                text.setFont(ComponentFactory.getSystemFont());
                text.setForeground(ComponentFactory.getDisabledColor());
                columnNew.setCellEditor(new DefaultCellEditor(text));
                DCTableCellRenderer renderer = new DCTableCellRenderer();
                renderer.setFont(ComponentFactory.getSystemFont());
                columnNew.setCellRenderer(renderer);
            } else {
                switch (fieldTypes[i]) {
                    case ComponentFactory._NUMBERFIELD:
                        columnNew.setCellEditor(new DefaultCellEditor((JTextField) comp));
                        columnNew.setMaxWidth(100);
                        columnNew.setCellRenderer(new NumberTableCellRenderer());
                        break;
                    case ComponentFactory._LONGTEXTFIELD :
                    case ComponentFactory._SHORTTEXTFIELD :
                        columnNew.setCellEditor(new DefaultCellEditor((JTextField) comp));
                        break;
                    case ComponentFactory._TIMEFIELD :
                        TimeFieldTableCellRenderer timeRenderer = new TimeFieldTableCellRenderer();
                        NumberField numberField = ComponentFactory.getNumberField();
                        columnNew.setCellEditor(new DefaultCellEditor((JTextField) numberField));
                        columnNew.setCellRenderer(timeRenderer);
                        break;
                    case ComponentFactory._URLFIELD :
                        columnNew.setCellEditor(new DefaultCellEditor((JTextField) comp));
                        DCTableCellRenderer renderer = new DCTableCellRenderer();
                        renderer.setForeground(new Color(0, 0, 255));
                        columnNew.setCellRenderer(renderer);
                        break;
                    case ComponentFactory._PICTUREFIELD :
                        text = ComponentFactory.getTextField();
                        text.setEditable(false);
                        text.setFont(ComponentFactory.getUnreadableFont());
                        columnNew.setCellEditor(new DefaultCellEditor(text));
                        columnNew.setCellRenderer(
                                new PictureTableCellRenderer(module.getIndex()));
                        break;                        
                    case ComponentFactory._BOOKCATEGORYCOMBO:                        
                    case ComponentFactory._SOFTWARECATEGORYCOMBO:
                        columnNew.setCellRenderer(new ComboBoxTableCellRenderer());
                        columnNew.setCellEditor(new DefaultCellEditor((JComboBox) comp));
                        break;
                    case ComponentFactory._RATINGCOMBOBOX:
                        columnNew.setMinWidth(70);
                        columnNew.setCellRenderer(new RatingTableCellRenderer());
                        columnNew.setCellEditor(new DefaultCellEditor((RatingComboBox) comp));
                        break;
                    case ComponentFactory._MP3GENRECOMBO:
                    case ComponentFactory._YEARCOMBO:
                    case ComponentFactory._YESNOCOMBO:
                        columnNew.setCellEditor(new DefaultCellEditor((JComboBox) comp));
                        break;
                }    
                comp.setEnabled(!readonly);
            }
        }

        createRowIndexColumn();
        addRowIndexColumn();
        setProperties();        
    }
    
    /**
     * Adds all the columns to the table (hidden and visible fields)
     */
    private void resetTable() {
        for (Iterator iter = alColumns.iterator(); iter.hasNext(); ) {
            TableColumn column = (TableColumn) iter.next();
            try {
                getColumn(column.getIdentifier());
            } catch (Exception exp) {
                addColumn(column);
            }
        }
    }      
    
    /**
     * Applies all default properties on the table
     */
    private void setProperties() {
        setBackground(ComponentFactory.getTableColor());
        setAutoscrolls(true);

        setDefaultRenderer(Object.class, new DCTableCellRenderer());
        setFont(ComponentFactory.getStandardFont());

        setRowSelectionAllowed(true);
        setColumnSelectionAllowed(false);
        setRequestFocusEnabled(true);
        setAutoResizeMode(JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS);
        setAlignmentY(JTable.TOP_ALIGNMENT);
        getTableHeader().setReorderingAllowed(true);
        setShowHorizontalLines(true);
        
        setBackground(new Color(255, 255, 255));
        setGridColor(new Color(220,220,200));
        
        applyHeaders();

        setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
    }    
    
    /**
     * Enables / disables caching of changed values
     * @param b
     */
    private void setListeningForChanges(boolean b) {
        boolean enable = b && caching;
        if (enable && getListeners(tableChangeListener.getClass()).length == 0) {
            tableModel.addTableModelListener(tableChangeListener);
        } else if (!enable) {
            tableModel.removeTableModelListener(tableChangeListener);
        }
    }
    
    private DefaultTableCellRenderer getRequiredTableHeaderRenderer() {
        DCTableHeaderRendererRequired hdrRenderer = new DCTableHeaderRendererRequired();
        return hdrRenderer;
    }

    private DefaultTableCellRenderer getTableHeaderRenderer() {
        DCTableHeaderRenderer hdrRenderer = new DCTableHeaderRenderer();
        return hdrRenderer;
    }
    
    private void createRowIndexColumn() {
        rowIndexColumn = new TableColumn();
        rowIndexColumn.setHeaderValue("Row");
        rowIndexColumn.setMaxWidth(40);
        rowIndexColumn.setPreferredWidth(40);
        rowIndexColumn.setWidth(40);
        JTextField text = ComponentFactory.getTextField();
        text.setEditable(false);
        text.setEnabled(false);
        text.setBackground(DcSettings.getValueAsColor(Repository.settings.stColor));
        rowIndexColumn.setCellEditor(new DefaultCellEditor(text));
        DCTableRowIndexCellRenderer renderer = new DCTableRowIndexCellRenderer();
        renderer.setFont(ComponentFactory.getSystemFont());
        rowIndexColumn.setCellRenderer(renderer);
        rowIndexColumn.setHeaderRenderer(getTableHeaderRenderer());
    }
    
    private void addRowIndexColumn() {
        getColumnModel().addColumn(rowIndexColumn);
        moveColumn(getColumnCount() - 1, 0);
    }
    
    private void removeRowIndexColumn() {
    	getColumnModel().removeColumn(rowIndexColumn);
    }
    
    /**
     * Handles changes made to the table (on screen)
     */
    private class TableValueChangedAction implements TableModelListener {
        public void tableChanged(TableModelEvent e) {
            if (    !ignoreEdit &&
            		(e.getType() == TableModelEvent.INSERT || 
                     e.getType() == TableModelEvent.UPDATE)) {
            	
                int row = getSelectedRow();
                addRowToCache(row, e.getColumn());
            }
        }
    }    
}
