package net.sf.dc.reporting;

import java.awt.Color;
import java.io.File;
import java.io.FileOutputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import javax.swing.ImageIcon;
import javax.swing.JComboBox;
import javax.swing.JComponent;

import net.sf.dc.console.components.ComponentFactory;
import net.sf.dc.console.reporting.ReportingDialog;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.Repository;
import net.sf.dc.core.objects.DCFieldDefinition;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.core.objects.reporting.ReportFieldDefinition;
import net.sf.dc.core.objects.reporting.ReportFieldDefinitions;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.settings.ReportSettings;
import net.sf.ext.util.Resources;
import net.sf.ext.util.Utilities;

import com.lowagie.text.BadElementException;
import com.lowagie.text.Cell;
import com.lowagie.text.Chunk;
import com.lowagie.text.Document;
import com.lowagie.text.Font;
import com.lowagie.text.FontFactory;
import com.lowagie.text.HeaderFooter;
import com.lowagie.text.Image;
import com.lowagie.text.Paragraph;
import com.lowagie.text.Phrase;
import com.lowagie.text.Table;
import com.lowagie.text.pdf.PdfWriter;

public abstract class Report extends Thread {

    protected final Color textColor = ReportSettings.getValueAsColor(Repository.reportsettings.stTextColor);
    protected final Color textColorBg = ReportSettings.getValueAsColor(Repository.reportsettings.stTextBackgroundcolor);
    protected final Font textFont = getFont(ReportSettings.getValueAsFont(Repository.reportsettings.stTextFont), textColor);
    
    protected final Color textBorderColor = ReportSettings.getValueAsColor(Repository.reportsettings.stTextBorderColor);
    protected final Color labelBorderColor = ReportSettings.getValueAsColor(Repository.reportsettings.stLabelBorderColor);
    
    protected final Color labelColor = ReportSettings.getValueAsColor(Repository.reportsettings.stLabelColor);
    protected final Color labelColorBg = ReportSettings.getValueAsColor(Repository.reportsettings.stLabelBackgroundColor);
    protected final Font labelFont = getFont(ReportSettings.getValueAsFont(Repository.reportsettings.stLabelFont), labelColor);
    
    protected final int textBorderBottom = ReportSettings.getValueAsBoolean(Repository.reportsettings.stTextBorderBottom) ? 1 : 0;        
    protected final int textBorderTop = ReportSettings.getValueAsBoolean(Repository.reportsettings.stTextBorderTop) ? 1 : 0;
    protected final int textBorderLeft = ReportSettings.getValueAsBoolean(Repository.reportsettings.stTextBorderLeft) ? 1 : 0;
    protected final int textBorderRight = ReportSettings.getValueAsBoolean(Repository.reportsettings.stTextBorderRight) ? 1 : 0;

    protected final int labelBorderBottom = ReportSettings.getValueAsBoolean(Repository.reportsettings.stLabelBorderBottom) ? 1 : 0;        
    protected final int labelBorderTop = ReportSettings.getValueAsBoolean(Repository.reportsettings.stLabelBorderTop) ? 1 : 0;
    protected final int labelBorderLeft = ReportSettings.getValueAsBoolean(Repository.reportsettings.stLabelBorderLeft) ? 1 : 0;
    protected final int labelBorderRight = ReportSettings.getValueAsBoolean(Repository.reportsettings.stLabelBorderRight) ? 1 : 0;
    
    
    protected ReportFieldDefinitions definitions;
    protected DCObject[] data;
    protected File target;
    protected ReportingDialog dialog;

    protected Document document = new Document();
    protected boolean keepOnRunning = true;

    public void run() {
        try {
            create();
        } catch (Exception exp) {
            document.close();
            dialog.addMessage(Resources.getText(DCMessages.msgErrorWhileCreatingReport, 
                          exp.toString()));
        } finally {
            data = null;
            dialog.allowActions(true);   
        }            
    }

    public void cancel() {
        keepOnRunning = false;
    }    
    
    public void compile() {
        super.start();
    }

    
    public abstract void create() throws Exception;    
    
    
    protected void initialize() throws Exception {
        document.addCreator("Data Crow");
        document.addAuthor("Robert Jan van der Waals");
        FileOutputStream out = new FileOutputStream(target);

        PdfWriter.getInstance(document, out);
        
        document.open();
        if (ReportSettings.getValueAsBoolean(Repository.reportsettings.stShowFooter))
            document.setFooter(getDocumentFooter());

        if (ReportSettings.getValueAsBoolean(Repository.reportsettings.stShowHeader))
            document.setHeader(getDocumentHeader());
        
        addFrontPage(document);
        document.newPage();
        
        dialog.initProgressBar(data.length);
        dialog.addMessage(Resources.getText(DCMessages.msgCreatingReportForXObjects, 
                                            String.valueOf(data.length)));
    }
    
    protected Font getFont(java.awt.Font f, Color color) {
        return FontFactory.getFont(f.getName(), f.getSize(), f.getStyle(), color);
    }
    
    protected Table getTable(int columnCount) throws BadElementException {
        Table datatable = new Table(columnCount);
        datatable.setSpaceInsideCell(1f);
        datatable.setBorder(0);
        datatable.setWidth(100f);
        return datatable;
    }     
    
    private int getAlignment(ReportFieldDefinition definition) {
        if (definition != null) {
            int alignment = definition.getAlignment();
            
            if (alignment == ReportFieldDefinition._CENTER) {
                return Cell.ALIGN_CENTER;
            } else if (alignment == ReportFieldDefinition._LEFT) {
                return Cell.ALIGN_LEFT;
            } else {
                return Cell.ALIGN_RIGHT;
            }
        } else {
            return Cell.ALIGN_LEFT;
        }
    }
    
    protected Cell getLabelCell(String label, ReportFieldDefinition definition) throws BadElementException {
        Chunk chunk = new Chunk(label);
        Cell cell = new Cell(chunk);
        chunk.setFont(labelFont);

        cell.setHorizontalAlignment(getAlignment(definition));
        
        cell.setBackgroundColor(labelColorBg);
        
        cell.setBorderWidthBottom(labelBorderBottom);
        cell.setBorderWidthLeft(labelBorderLeft);
        cell.setBorderWidthRight(labelBorderRight);
        cell.setBorderWidthTop(labelBorderTop);

        cell.setBorderColorBottom(labelBorderColor);
        cell.setBorderColorLeft(labelBorderColor);
        cell.setBorderColorRight(labelBorderColor);
        cell.setBorderColorTop(labelBorderColor);
        return cell;
    }

    protected Cell getTextCell(String text, ReportFieldDefinition definition) throws BadElementException {
        Chunk chunk = new Chunk(text);
        Cell cell = new Cell(chunk);
        chunk.setFont(textFont);
        setTextCellDefaults(cell, definition);
        return cell;
    }  

    protected Cell getRatingCell(Object value, ReportFieldDefinition definition) throws Exception {
        ImageIcon icon = Utilities.getRatingIcon(value);
        Image image = Image.getInstance(icon.getImage(), null, false);
        Cell cell = new Cell(image);
        setTextCellDefaults(cell, definition);
        return cell;
    }
    
    protected Cell getImageCell(URL url, ReportFieldDefinition definition, int width, int height, boolean inRow) throws Exception {
        ImageIcon icon = new ImageIcon(url);
        Image image = Image.getInstance(icon.getImage(), null, false);
        
        if (icon.getIconWidth() > width || icon.getIconHeight() > height)
            image.scaleToFit(width, height);
        
        Cell cell = new Cell(image);
        cell.setBorder(0);
        
        if (inRow) setTextCellDefaults(cell, definition);
        
        return cell;
    }
    
    protected void setWidths(Table table, DCObject dco, ReportFieldDefinitions definitions) {
        Collection fields = new ArrayList();
        for (Iterator iter = definitions.iterator(); iter.hasNext(); ) {
            ReportFieldDefinition definition = (ReportFieldDefinition) iter.next();
            if (definition.isEnabled()) {
                fields.add(dco.getField(definition.getIndex()));
            }
        }
        
        int[] widths = new int[fields.size()];
        int i = 0;
        for (Iterator iter = fields.iterator(); iter.hasNext(); i++) {
            DCFieldDefinition field = (DCFieldDefinition) iter.next();
            
            int fieldtype = field.getFieldType();
            int valuetype = field.getValueType();
            JComponent comp = ComponentFactory.getComponent(fieldtype, "", 100);
            if (fieldtype == ComponentFactory._RATINGCOMBOBOX ||
                fieldtype == ComponentFactory._PICTUREFIELD ||
                comp instanceof JComboBox ||
                valuetype == Repository.datatypes._INTEGER ||
                field.getIndex() > 5) { // TODO : unacceptable hack ;)
                
                widths[i] = (int) ((100 / fields.size())); 
            } else {
                widths[i] = (100 / fields.size()) * 3;
            }
        }
        
        try {
            table.setWidth(100f);
            table.setWidths(widths);
        } catch (Exception exp) {
            Messages.add(exp);
        }
    }
    
    private void setTextCellDefaults(Cell cell, ReportFieldDefinition definition) {
        
        cell.setHorizontalAlignment(getAlignment(definition));
        
        cell.setBackgroundColor(textColorBg);
        
        cell.setBorderWidthBottom(textBorderBottom);
        cell.setBorderWidthLeft(textBorderLeft);
        cell.setBorderWidthRight(textBorderRight);
        cell.setBorderWidthTop(textBorderTop);
        
        cell.setBorderColorBottom(textBorderColor);
        cell.setBorderColorLeft(textBorderColor);
        cell.setBorderColorRight(textBorderColor);
        cell.setBorderColorTop(textBorderColor);   
    }    
    
    private HeaderFooter getDocumentFooter() {
        Color fontColor = ReportSettings.getValueAsColor(Repository.reportsettings.stFooterColor);
        Font font = getFont(ReportSettings.getValueAsFont(Repository.reportsettings.stFooterFont), fontColor);
        String value = ReportSettings.getValueAsString(Repository.reportsettings.stFooterValue);
        
        Chunk chunk1 = new Chunk(value + " ");
        chunk1.setFont(font);
        Chunk chunk2 = new Chunk("");
        chunk2.setFont(font);
        
        HeaderFooter footer = new HeaderFooter(new Phrase(chunk1), 
                                               new Phrase(chunk2));
        footer.setBorder(0);
        Color background = ReportSettings.getValueAsColor(Repository.reportsettings.stFooterBackgroundColor);
        footer.setBackgroundColor(background);
        return footer;
    }
    
    private HeaderFooter getDocumentHeader() {
        String text = ReportSettings.getValueAsString(Repository.reportsettings.stHeaderValue);
        Chunk chunk = new Chunk(text);
        
        Color fontColor = ReportSettings.getValueAsColor(Repository.reportsettings.stHeaderColor);
        Font font = getFont(ReportSettings.getValueAsFont(Repository.reportsettings.stHeaderFont), fontColor);
        chunk.setFont(font);
        HeaderFooter header = new HeaderFooter(new Phrase(chunk), false);
        header.setBorder(0);
        
        Color background = ReportSettings.getValueAsColor(Repository.reportsettings.stHeaderBackgroundColor);
        header.setBackgroundColor(background);
        return header;
    }    
    
    private Paragraph addFrontPage(Document document) throws Exception {
        Paragraph paragraph = new Paragraph();
        paragraph.setAlignment(Paragraph.ALIGN_RIGHT);
        Chunk chunk = new Chunk("a data crow report");
        Font font = FontFactory.getFont("Arial", 8, Font.ITALIC, Color.BLACK);
        chunk.setFont(font);
        paragraph.add(chunk);
        document.add(paragraph);

        Paragraph imgParagraph = new Paragraph();
        imgParagraph.setAlignment(Paragraph.ALIGN_CENTER);

        Image image = Image.getInstance(IconLibrary._icoReportTitle.getImage(), null, false);
        Chunk imageChunk = new Chunk(image, -20f, -500f);
        imageChunk.setHorizontalScaling(0f);
        imgParagraph.add(imageChunk);
        document.add(imgParagraph);
        
        return paragraph;        
    }       
}
